#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Any, Final, Optional, Type, TypeVar, Union, Tuple, List, Dict, Set, cast
import builtins
import importlib.util
import os
# from .string import string


#--------------------------------------------------------------------------------
# 전역 상수 목록.
#--------------------------------------------------------------------------------
PYEXTENSION : str = "py"
INITNAME : str = "__init__"
INITFILENAME : str = f"{INITNAME}.{PYEXTENSION}"
BUILTIN : str = "built-in"
EMPTY : str = ""


#--------------------------------------------------------------------------------
# 노드.
#--------------------------------------------------------------------------------
class Node:
	#--------------------------------------------------------------------------------
	# 공개 멤버 변수 목록.
	#--------------------------------------------------------------------------------
	Name : str # 패키지, 모듈, 클래스, 함수 이름.
	Path : str # 패키지를 포함한 전체 이름.
	IsPackage : bool # 패키지 여부. 
	Parent : Node # 부모.
	Children : list # 자식들.

	#--------------------------------------------------------------------------------
	# 생성됨.
	#--------------------------------------------------------------------------------
	def __init__(self, name : str = EMPTY, path : str = EMPTY, parent : Node = None, isPackage : bool = False):
		self.Name = name
		self.Path = path
		self.IsPackage = isPackage
		self.Parent = parent		
		self.Children = list()

	#--------------------------------------------------------------------------------
	# 클래스 문자열 변환됨.
	#--------------------------------------------------------------------------------
	def __repr__(self, level : int = 0):
		ret = "\t" * level + repr(self.Name) + "\n"
		for child in self.Children:
			ret += child.__repr__(level + 1)
		return ret
	
	#--------------------------------------------------------------------------------
	# 자식 노드 추가.
	#--------------------------------------------------------------------------------
	def AddChild(self, childNode : Node):
		childNode.Parent = self
		self.Children.append(childNode)

	#--------------------------------------------------------------------------------
	# 패키지 여부.
	# - 폴더 경로를 넣어서 확인.
	#--------------------------------------------------------------------------------
	@staticmethod
	def CheckPackage(path : str) -> bool:
		if not os.path.isdir(path):
			return False
		initFilePath = os.path.join(path, INITFILENAME)
		if not os.path.isfile(initFilePath):
			return False
		return True

	#--------------------------------------------------------------------------------
	# 폴더 기반 트리 구조 작성.
	#--------------------------------------------------------------------------------
	@staticmethod
	def BuildTree(path : str, parent : Node = None) -> Node:
		name = os.path.basename(path)
		node = Node(name, path, parent, True)
		for childName in os.listdir(path):
			childPath = os.path.join(path, childName)
			if os.path.isdir(childPath):
				if Node.CheckPackage(childPath):
					child = Node.BuildTree(childPath, node)
					node.AddChild(child)
			else:
				cpath, cname, cextension = str.GetSplitFilePath(childPath)				
				if not cextension.endswith(PYEXTENSION):
					continue
				if cname == INITNAME:
					continue
				child = Node(cname, childPath, parent, False)
				node.AddChild(child)
		return node

	#--------------------------------------------------------------------------------
	# 트리 탐색.
	#--------------------------------------------------------------------------------
	@staticmethod
	def TraverseTree(node : Node, prefix : str = EMPTY, usePrint : bool = True, moduleFullNames : dict[str, str] = None) -> None:
		if node.IsPackage:
			path = f"{prefix}.{node.Name}" if prefix else node.Name
			if usePrint:
				builtins.print(f"package: {path}")
			for child in node.Children:
				Node.TraverseTree(child, path, usePrint, moduleFullNames)
		else:
			path = f"{prefix}.{node.Name}" if prefix else node.Name
			if usePrint:
				builtins.print(f"module: {path}")
			if not moduleFullNames is None:
				moduleFullNames[path] = node.Name

	#--------------------------------------------------------------------------------
	# 노드 기준 모듈 이름 반환.
	# - KEY : 패키지포함모듈명, VALUE : 모듈명
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetModuleNames(node : Node) -> dict[str, str]:
		moduleFullNames = dict()
		Node.TraverseTree(node, EMPTY, False, moduleFullNames)
		return moduleFullNames


#--------------------------------------------------------------------------------
# 실제 존재하는 패키지 혹은 모듈 인지 여부.
#--------------------------------------------------------------------------------
def IsExistsPackageOrModule(packageOrModuleName : str) -> bool:
	try:
		# 모듈 정보 가져오기.
		moduleSpec = importlib.util.find_spec(packageOrModuleName)
		if not moduleSpec:
			return False
		else:
			# 모듈이 존재할 경우 origin은 해당 모듈의 경로 혹은 인터프리터 내장 라이브러리 식별자.
			moduleFilePath = moduleSpec.origin
			if not moduleFilePath:
				return False
			# 빌트인 인 경우는 실제 파일은 없지만 파이썬 인터프리터에 내장된 라이브러리이므로 있다고 간주.
			if not moduleFilePath == BUILTIN:
				return True
			# 경로가 실존하는 파일이나 폴더일 경우 있다고 간주.
			if os.path.isfile(moduleFilePath) or os.path.isdir(moduleFilePath):
				return True                
			return False
	except Exception as exception:
		# Application.LogException(exception, False, False)
		builtins.print(f"IsExistsPackageOrModule: {exception}")
		return False


#--------------------------------------------------------------------------------
# 실제 존재하는 어트리뷰트인지 여부.
#--------------------------------------------------------------------------------
def IsExistsAttribute(moduleName : str, attributeName : str) -> bool:
	if not IsExistsPackageOrModule(moduleName):
		return False
	try:
		module = importlib.import_module(moduleName)
		if not builtins.hasattr(module, attributeName):
			return False
	except Exception as exception:
		builtins.print(f"IsExistsAttribute: {exception}")
		return False
	return True


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Any, Final, Optional, Type, TypeVar, Union, Tuple, List, Dict, Set, cast
import builtins
from .jsonutility import STRToJSON, GetPrettyJSONString, RemoveAllCommentsInString
from .moduleutility import Node, IsExistsPackageOrModule, IsExistsAttribute
from .strutility import GetSplitFilePath, GetStringFromSeperatedStringList, CreateStringListFromSeperatedStringList, GetTimestampString


#--------------------------------------------------------------------------------
# 공개 클래스 목록.
#--------------------------------------------------------------------------------
__all__ = [

	# jsonutility.
	"STRToJSON",
	"GetPrettyJSONString",
	"RemoveAllCommentsInString",

	# moduleutility.
	"Node",
	"IsExistsPackageOrModule",
	"IsExistsAttribute",

	# strutility.
	"GetSplitFilePath",
	"GetStringFromSeperatedStringList",
	"CreateStringListFromSeperatedStringList",
	"GetTimestampString"
]


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Any, Final, Optional, Type, TypeVar, Union, Tuple, List, Dict, Set, cast
import builtins
import os
from datetime import datetime as DateTime


#--------------------------------------------------------------------------------
# 전역 상수 목록.
#--------------------------------------------------------------------------------
EMPTY : str = ""
NONE : str = "NONE"
COLON : str = ":"
SPACE : str = " "
SLASH : str = "/"
HYPHEN : str = "-"
COMMA : str = ","


#--------------------------------------------------------------------------------
# 파일의 전체경로를 파일의 (경로, 이름, 확장자) 분리.
#--------------------------------------------------------------------------------
def GetSplitFilePath(filePath : str, useLowerExtension : bool = True) -> tuple[str, str, str]:
	path, name = os.path.split(filePath)
	name, extension = os.path.splitext(name)
	if useLowerExtension:
		extension = extension.lower()
	return path, name, extension


#--------------------------------------------------------------------------------
# 구분자로 구분된 스트링을 하나의 리스트로 변환한다. 
# - 예 : "A/B/C" ==> ["A", "B", "C"]
#--------------------------------------------------------------------------------
def GetStringFromSeperatedStringList(string : str, seperator : str = COMMA) -> list[str]:
	if not string:
		return list()

	if seperator in string:
		result = string.split(seperator)
	else:
		result = list()
		result.append(string)

	return result

#--------------------------------------------------------------------------------
# 구분자로 구분된 스트링의 리스트를 하나의 리스트로 통합한다. 
# - 예 : ["A,B,C", "A,B,C", "A,B,C", ...] ==> ["A,B,C","A,B,C","A,B,C"]
#--------------------------------------------------------------------------------
def CreateStringListFromSeperatedStringList(strings : list[str], seperator : str = COMMA) -> list[str]:
	if not strings:
		return list()	
	result = list()
	for string in strings:
		if seperator in string: result.extend(filter(None, string.split(seperator)))
		else: result.append(string)
	return result


#--------------------------------------------------------------------------------
# 현재 시간 텍스트 반환.
#--------------------------------------------------------------------------------
def GetTimestampString(seperateDate : str = EMPTY, seperateGap : str = SPACE, seperateTime : str = EMPTY, useMilliseconds : bool = False, seperateMillySecondsTime : str = COMMA) -> str:
	nowTime : DateTime = DateTime.now()
	if useMilliseconds:
		milliseconds = nowTime.strftime("%f")[:3]
		timestamp = nowTime.strftime(f"%Y{seperateDate}%m{seperateDate}%d{seperateGap}%H{seperateTime}%M{seperateTime}%S{seperateMillySecondsTime}{milliseconds}")
	else:
		timestamp = nowTime.strftime(f"%Y{seperateDate}%m{seperateDate}%d{seperateGap}%H{seperateTime}%M{seperateTime}%S")
	return timestamp


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Any, Final, Optional, Type, TypeVar, Union, Tuple, List, Dict, Set, cast
import builtins
import logging
import os
import queue
from logging import Logger, handlers, Handler, StreamHandler, FileHandler, Formatter, LogRecord
from .ansicode import ANSICODE
from .strutility import GetTimestampString

#--------------------------------------------------------------------------------
# 전역 상수 목록.
#--------------------------------------------------------------------------------
EMPTY : str = ""
NONE : str = "NONE"
COLON : str = ":"
SPACE : str = " "
SLASH : str = "/"
HYPHEN : str = "-"
COMMA : str = ","
UTF8 : str = "utf-8"



# #--------------------------------------------------------------------------------
# # 화면 출력 핸들러.
# #--------------------------------------------------------------------------------
# class PrintHandler(Handler):
# 	#--------------------------------------------------------------------------------
# 	# 호출됨.
# 	#--------------------------------------------------------------------------------
# 	def emit(self, record : LogRecord):
# 		message = self.format(record)
# 		# if record.levelno == FATAL or record.levelno == CRITICAL: Print(f"<bg_red><white><b>{message}</b></white></bg_red>")
# 		# elif record.levelno == ERROR: Print(f"<red>{message}</red>")
# 		# elif record.levelno == WARN or record.levelno == WARNING: Print(f"<yellow>{message}</yellow>")
# 		# elif record.levelno == INFO: Print(f"{message}")
# 		# elif record.levelno == DEBUG: Print(f"<magenta>{message}</magenta>")
# 		PrintLog(message, record.levelno)


#--------------------------------------------------------------------------------
# 로그 파일 클래스.
#--------------------------------------------------------------------------------
class LOGFile:
	#--------------------------------------------------------------------------------
	# 멤버 변수 목록.
	#--------------------------------------------------------------------------------
	__ansicode : ANSICODE


	#--------------------------------------------------------------------------------
	# 생성됨.
	#--------------------------------------------------------------------------------
	def __init__(self) -> None:
		self.__ansicode = ANSICODE()


	#--------------------------------------------------------------------------------
	# 기록 시작.
	# - 로그 파일 사용 설정을 비활성화 하면 로그는 CLI상에서만 출력된다.
	#--------------------------------------------------------------------------------
	def Run(loggerName : str, logLevel : int, useLogFile : bool, logPath : str):

		# logLevel : int = logging.NOTSET
		timestamp : str = GetTimestampString(EMPTY, EMPTY, EMPTY, True, EMPTY)
		logFilePath : str = f"{logPath}/{loggerName}-{timestamp}.log"

		# # EXE 파일 실행.
		# if Application.IsBuild():
		# 	useLogFile = False
		# 	logLevel = logging.WARNING
		# # VSCode에서 디버깅 실행.
		# elif Application.IsDebug():
		# 	useLogFile = True
		# 	logLevel = logging.DEBUG
		# 	logFilePath = Application.GetRootPathWithRelativePath(f"logs/pyappcore-debug-{timestamp}.log")
		# # Blender.exe로 소스코드 실행.
		# elif Application.HasSymbol(SYMBOL_SERVICE):
		# 	useLogFile = True
		# 	logLevel = logging.INFO
		# 	logFilePath = Application.GetRootPathWithRelativePath(f"logs/pyappcore-service-{timestamp}.log")
		# # VSCode에서 디버깅 없이 실행.
		# else:
		# 	useLogFile = True
		# 	logLevel = logging.INFO
		# 	logFilePath = Application.GetRootPathWithRelativePath(f"logs/pyappcore-nodebug-{timestamp}.log")


		# 설정.
		logger : Logger = logging.getLogger(loggerName)
		logger.setLevel(logLevel)

		# 로깅 큐 추가.
		# 로그 파일 기록이 자꾸 씹히는 이슈 있어서 사용. (하지만 개선 효과 없음)
		logQueue = queue.Queue()
		ququeHandler = handlers.QueueHandler(logQueue)
		logger.addHandler(ququeHandler)

		# 로그 출력 양식 설정.
		# formatter : Formatter = Formatter("[%(asctime)s][%(name)s][%(levelname)s] %(message)s")
		formatter : Formatter = Formatter("[%(asctime)s][%(levelname)s] %(message)s")

		# 프린트 핸들러.
		# printHandler : PrintHandler = PrintHandler()
		# printHandler.setLevel(logLevel)
		# printHandler.setFormatter(formatter)
		# applicationLogger.addHandler(printHandler)

		# 로그파일 설정.
		if useLogFile:
			if not os.path.isdir(logPath):
				os.makedirs(logPath)
			
			fileHandler : StreamHandler = FileHandler(logFilePath, encoding = UTF8)
			fileHandler.setLevel(logLevel)
			fileHandler.setFormatter(formatter)
			# applicationLogger.addHandler(fileHandler)
			# queueListener = handlers.QueueListener(logQueue, printHandler, fileHandler)

			# 큐 시작.
			queueListener = handlers.QueueListener(logQueue, fileHandler)
			queueListener.start()


	#--------------------------------------------------------------------------------
	# 로그 출력.
	#--------------------------------------------------------------------------------
	def PrintLog(self, text : str, logLevel : int) -> None:
		if logLevel == logging.FATAL or logLevel == logging.RITICAL: self.__ansicode.Print(f"<bg_red><white><b>{text}</b></white></bg_red>")
		elif logLevel == logging.ERROR: self.__ansicode.Print(f"<red>{text}</red>")
		elif logLevel == logging.WARN or logLevel == logging.WARNING: self.__ansicode.Print(f"<yellow>{text}</yellow>")
		elif logLevel == logging.INFO: self.__ansicode.Print(f"{text}")
		elif logLevel == logging.DEBUG: self.__ansicode.Print(f"<magenta>{text}</magenta>")
		else: self.__ansicode.Print(text)


	#--------------------------------------------------------------------------------
	# 로그 수준에 대한 문자열 출력.
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetStringFromLogLevel(logLevel : int) -> str:
		return logging.getLevelName(logLevel)